package toy.dialog;

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;

import toy.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyConverterPanel extends Panel implements KeyListener, TextListener
{
    public static final Font MONOSPACED_FONT = new Font("Monospaced", Font.PLAIN, 12);
    public static final Font SANSSERIF_FONT = new Font("SansSerif", Font.PLAIN, 12);
    public static final Font LABEL_FONT = new Font("SansSerif", Font.BOLD, 12);

    public static final int MIN_WIDTH = 190;
    public static final int HEIGHT = 105;
        //the 16-bit converterPanel won't look very good if it's smaller than this

    private ToyFocusManager focusManager;

    //The dialog items
    private Label binaryLabel;
    private TextField binaryField;
    private Label decimalLabel;
    private TextField decimalField;
    private Label hexLabel;
    private TextField hexField;
    private Label pseudoCodeLabel;
    private TextField pseudoCodeField;

    ToyWord baseConverter;
        //Although it seems strange to be using an instance of a toy memory
        //slot to perform base conversions, it's extremely useful

    public ToyConverterPanel(ToyFocusManager newFocusManager)
    {
        super(new GridBagLayout());

        focusManager= newFocusManager;

        setBackground(new Color(192, 192, 192));

        //Binary Label/Field
        binaryLabel = new Label("Bit:");
        binaryLabel.setFont(LABEL_FONT);
        binaryLabel.setForeground(new Color(0, 0, 0));
        add(
            binaryLabel,
            ToyToolkit.createGridBagConstraints(
                0, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.WEST,
                GridBagConstraints.NONE,
                new Insets(5, 5, 5, 5),
                0, 0
                )
            );

        binaryField = new TextField(16);
        binaryField.addFocusListener(focusManager);
        binaryField.addKeyListener(this);
        binaryField.addTextListener(this);
        binaryField.setBackground(new Color(255, 255, 255));
        binaryField.setFont(MONOSPACED_FONT);
        binaryField.setForeground(new Color(0, 0, 0));
        add(
            binaryField,
            ToyToolkit.createGridBagConstraints(
                1, 0,
                1, 1,
                100.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(5, 0, 5, 5),
                0, 0
                )
            );

        //Decimal Label/Field
        decimalLabel = new Label("Dec:");
        decimalLabel.setFont(LABEL_FONT);
        decimalLabel.setForeground(new Color(0, 0, 0));
        add(
            decimalLabel,
            ToyToolkit.createGridBagConstraints(
                0, 1,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.WEST,
                GridBagConstraints.NONE,
                new Insets(0, 5, 5, 5),
                0, 0
                )
            );

        decimalField = new TextField(10);
        decimalField.addFocusListener(focusManager);
        decimalField.addKeyListener(this);
        decimalField.addTextListener(this);
        decimalField.setBackground(new Color(255, 255, 255));
        decimalField.setFont(MONOSPACED_FONT);
        decimalField.setForeground(new Color(0, 0, 0));
        add(
            decimalField,
            ToyToolkit.createGridBagConstraints(
                1, 1,
                1, 1,
                100.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(0, 0, 5, 5),
                0, 0
                )
            );

        //Hexidecimal Label/Field
        hexLabel = new Label("Hex:");
        hexLabel.setFont(LABEL_FONT);
        hexLabel.setForeground(new Color(0, 0, 0));
        add(
            hexLabel,
            ToyToolkit.createGridBagConstraints(
                0, 2,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.WEST,
                GridBagConstraints.NONE,
                new Insets(0, 5, 5, 5),
                0, 0
                )
            );

        hexField = new TextField(4);
        hexField.addFocusListener(focusManager);
        hexField.addKeyListener(this);
        hexField.addTextListener(this);
        hexField.setBackground(new Color(255, 255, 255));
        hexField.setFont(MONOSPACED_FONT);
        hexField.setForeground(new Color(0, 0, 0));
        add(
            hexField,
            ToyToolkit.createGridBagConstraints(
                1, 2,
                1, 1,
                100.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(0, 0, 5, 5),
                0, 0
                )
            );

        //PseudoCode Label/Field
        pseudoCodeLabel = new Label("Pse:");
        pseudoCodeLabel.setFont(LABEL_FONT);
        pseudoCodeLabel.setForeground(new Color(0, 0, 0));
        add(
            pseudoCodeLabel,
            ToyToolkit.createGridBagConstraints(
                0, 3,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.WEST,
                GridBagConstraints.NONE,
                new Insets(0, 5, 5, 5),
                0, 0
                )
            );

        pseudoCodeField = new TextField(25);
        pseudoCodeField.addFocusListener(focusManager);
        pseudoCodeField.setBackground(new Color(192, 193, 192));
        pseudoCodeField.setEditable(false);
        pseudoCodeField.setFont(MONOSPACED_FONT);
        pseudoCodeField.setForeground(new Color(0, 0, 0));
        add(
            pseudoCodeField,
            ToyToolkit.createGridBagConstraints(
                1, 3,
                1, 1,
                100.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(0, 0, 5, 5),
                0, 0
                )
            );


        //change the converter's data
        baseConverter = ToyWord.parseDecimal(0);
        binaryField.setText("000000000000000");
        decimalField.setText("0");
        hexField.setText("0000");
    }

    ////////////////////////////////////////////////////////////////////////////
    // intercept setBounds() calls
    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        super.setBounds(x, y, width, height);

        doLayout();
    }

    public Dimension getMaximumSize()
    {
        return super.getPreferredSize();
    }

    //--------------------------------------------------------------------------
    //Update the Dec and Hex field
    //--------------------------------------------------------------------------
    private void binaryFieldChanged()
    {
        boolean valid = true;
        boolean bits[] = new boolean[16];
        int length = binaryField.getText().length();
        int temp;

        //are we in the proper length limits?
        if (length == 0)
            valid = false;
        if (length > 16)
        {
            for (int ctr = 0; valid && ctr < length - 16; ctr++)
                if (binaryField.getText().charAt(ctr) != '0')
                    valid = false;
        }

        //are each of our characters valid, and what is the decimal value?
        for (int ctr = 0; valid && ctr < length && ctr < 16; ctr++)
        {
            temp = binaryField.getText().charAt(length - ctr - 1) - '0';
            if (temp == 1 || temp == 0)
                bits[ctr] = (temp == 1);
            else
                valid = false;
        }

        if (valid)
        {   //load the new number into baseConverter and change the other fields
            baseConverter = ToyWord.parseBinary(bits);

            //System.out.println("Bit changed, updating others");
            decimalField.setText(baseConverter.toDecimalString());
            hexField.setText(baseConverter.toHexString());
            pseudoCodeField.setText(baseConverter.toPseudoCodeString());

            return;
        }
        else
        {   //make the base converter undefined and clear the other fields
            baseConverter = ToyWord.UNDEFINED;

            //System.out.println("Bit incorrect, clearing others");
            decimalField.setText("");
            hexField.setText("");
            pseudoCodeField.setText("");
            return;
        }
    }

    //--------------------------------------------------------------------------
    //Uppercase the Dec field and update the Hex field
    //--------------------------------------------------------------------------
    private void decimalFieldChanged()
    {
        boolean valid = true;
        int dec = 0;
        int length = decimalField.getText().length();
        char temp;

        //are we in the proper length limits?
        if (length == 0)
            valid = false;

        //are each of our characters valid, and what is the decimal value?
        for (int ctr = 0; valid && ctr < length; ctr++)
        {
            temp = decimalField.getText().charAt(length - ctr - 1);

            if (temp >= '0' && temp <= '9')
                dec += (temp - '0') * (int)Math.pow(10, ctr);
                else
                {
                    if (temp == '-' && ctr == length - 1)
                        dec *= -1;
                        else
                        valid = false;
                }

            if (valid && (dec > Short.MAX_VALUE || dec < Short.MIN_VALUE))
                valid = false;
        }

        if (valid)
        {   //load the new number into baseConverter and change the other fields
            baseConverter = ToyWord.parseDecimal(dec);

            //System.out.println("Dec changed, updating others");
            hexField.setText(baseConverter.toHexString());
            binaryField.setText(baseConverter.toBinaryString());
            pseudoCodeField.setText(baseConverter.toPseudoCodeString());
            return;
        }
        else
        {   //make the base converter undefined and clear the other fields
            baseConverter = ToyWord.UNDEFINED;

            //System.out.println("Dec invalid, clearing others");
            hexField.setText("");
            binaryField.setText("");
            pseudoCodeField.setText("");
            return;
        }
    }

    //--------------------------------------------------------------------------
    //Uppercase the Hex field and update the Dec field
    //--------------------------------------------------------------------------
    private void hexFieldChanged()
    {
        boolean valid = true;
        char hex[] = {'0', '0', '0', '0'};
        int length = hexField.getText().length();
        char temp;


        //are we in the proper length limits?
        if (length == 0)
            valid = false;
        if (length > 4)
        {
            for (int ctr = 0; valid && ctr < length - 4; ctr++)
                if (hexField.getText().charAt(ctr) != '0')
                    valid = false;
        }

        //are each of our characters valid (while we're answering that question,
        //might as well read the value)?
        for (int ctr = Math.max(length - 4, 0); valid && ctr < length; ctr++)
        {
            temp = hexField.getText().charAt(ctr);
            if (ToyWord.isHexDigit(temp))
                hex[ctr + 4 - length] = temp;
            else
                valid = false;
        }

        if (valid)
        {   //load the new number into baseConverter and change the other fields
            baseConverter = ToyWord.parseHexidecimal(hex[0], hex[1], hex[2], hex[3]);

            //System.out.println("Hex changed, updating others");
            decimalField.setText(baseConverter.toDecimalString());
            binaryField.setText(baseConverter.toBinaryString());
            pseudoCodeField.setText(baseConverter.toPseudoCodeString());
            return;
        }
        else
        {   //make the base converter undefined and clear the other fields
            baseConverter = ToyWord.UNDEFINED;

            //System.out.println("Hex invalid, clearing others");
            decimalField.setText("");
            binaryField.setText("");
            pseudoCodeField.setText("");
            return;
        }
    }

    //--------------------------------------------------------------------------
    //Implement KeyListener
    //--------------------------------------------------------------------------
    public void keyPressed(KeyEvent e)
    {

    }

    public void keyReleased(KeyEvent e)
    {
        //If Ctrl-z is pressed, IE will undo the typing but won't
        //trigger the textValueChanged event, so we'll do it manually
        if (e.getKeyCode() == java.awt.event.KeyEvent.VK_Z &&
               ((e.getModifiers() & KeyEvent.ALT_MASK)  != 0 ||
                (e.getModifiers() & KeyEvent.CTRL_MASK) != 0 ||
                (e.getModifiers() & KeyEvent.META_MASK) != 0))
            textValueChanged(new TextEvent(e.getSource(), TextEvent.TEXT_VALUE_CHANGED));
    }

    public void keyTyped(KeyEvent e)
    {

    }

    //--------------------------------------------------------------------------
    //Implement TextListener
    //--------------------------------------------------------------------------
    public void textValueChanged(TextEvent e)
    {
        if (e.getSource() == binaryField && focusManager.getFocusedComponent() == binaryField)
            binaryFieldChanged();
        if (e.getSource() == decimalField && focusManager.getFocusedComponent() == decimalField)
            decimalFieldChanged();
        if (e.getSource() == hexField && focusManager.getFocusedComponent() == hexField)
            hexFieldChanged();
    }
}