package toy.dialog;

import java.awt.*;
import java.awt.event.*;
import toy.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyEditPanel extends ToyRefreshablePanel
                          implements ActionListener, ItemListener, KeyListener
{
    ////////////////////////////////////////////////////////////////////////////
    // constants
    public static final Font MONOSPACED_FONT = new Font("Monospaced", Font.PLAIN, 12);
    public static final Font SANSSERIF_FONT = new Font("SansSerif", Font.PLAIN, 12);
    public static final Font LABEL_FONT = new Font("SansSerif", Font.PLAIN, 14);

    ////////////////////////////////////////////////////////////////////////////
    // variables
    private ToyWorkspace workspace;
    private ToyWarningHash warningHash;

    private ToyFocusManager focusManager;

    private CheckboxMenuItem autoCommentCheckbox;
    private CheckboxMenuItem highlightingCheckbox;
    private CheckboxMenuItem warningCheckbox;

    private MenuItem undoMenuItem;
    private MenuItem redoMenuItem;

    private boolean backspaceUndoableAutoComment;

    private MenuItem insertFunctionCommentMenuItem;
    private MenuItem appendCommentMenuItem;
    private MenuItem reformatCodeMenuItem;


    private Panel leftPanel;
        private Label programLabel;

        private Panel programPanel;
            private ToyTextArea programTextArea;
            private ToyTextAreaPeer programTextAreaPeer;

        private Label warningLabel;
        private List warningList;

    private Panel rightPanel;
        private Label inputLabel;
        private ToyInputPanel inputPanel;

        private Label referenceLabel;
        private TextArea referenceTextArea;

        private Label converterLabel;
        private ToyConverterPanel converterPanel;


    ////////////////////////////////////////////////////////////////////////////
    // the constructor
    public ToyEditPanel(ToyFocusManager newFocusManager,
                        CheckboxMenuItem newAutoCommentCheckbox,
                        CheckboxMenuItem newHighlightingCheckbox,
                        CheckboxMenuItem newWarningCheckbox,
                        MenuItem newUndoMenuItem,
                        MenuItem newRedoMenuItem,
                        MenuItem newInsertFunctionCommentMenuItem,
                        MenuItem newAppendCommentMenuItem,
                        MenuItem newReformatCodeMenuItem)
    {
        super(new GridBagLayout());

        focusManager = newFocusManager;
        autoCommentCheckbox = newAutoCommentCheckbox;
        highlightingCheckbox = newHighlightingCheckbox;
            highlightingCheckbox.addItemListener(this);
        warningCheckbox = newWarningCheckbox;
            warningCheckbox.addItemListener(this);

        undoMenuItem = newUndoMenuItem;
            undoMenuItem.addActionListener(this);
        redoMenuItem = newRedoMenuItem;
            redoMenuItem.addActionListener(this);
        insertFunctionCommentMenuItem = newInsertFunctionCommentMenuItem;
            insertFunctionCommentMenuItem.addActionListener(this);
        appendCommentMenuItem = newAppendCommentMenuItem;
            appendCommentMenuItem.addActionListener(this);
        reformatCodeMenuItem = newReformatCodeMenuItem;
            reformatCodeMenuItem.addActionListener(this);


        setBackground(new Color(80, 100, 120));

        leftPanel = new Panel(new GridBagLayout());
        leftPanel.setBackground(getBackground());
        add(
            leftPanel,
            ToyToolkit.createGridBagConstraints(
                0, 0,
                1, 1,
                1.5, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5),
                0, 0
                )
            );
        {
            programLabel = new Label();
            programLabel.setFont(LABEL_FONT);
            leftPanel.add(
                programLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 0,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            programPanel = new Panel(new GridLayout(1, 1));
            leftPanel.add(
                programPanel,
                ToyToolkit.createGridBagConstraints(
                    0, 1,
                    1, 1,
                    1.0, 60.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
            {
                programTextArea = new ToyTextArea(25, 90);
                programTextArea.addFocusListener(focusManager);
                programTextArea.addKeyListener(this);
                programTextArea.setBackground(new Color(255, 255, 255));
                programTextArea.setFont(MONOSPACED_FONT);
                programTextArea.setForeground(new Color(0, 0, 0));

                programTextAreaPeer = new ToyTextAreaPeer(programTextArea);
                programTextAreaPeer.setVisible(highlightingCheckbox.getState());

                if (highlightingCheckbox.getState())
                {
                    programPanel.add(programTextAreaPeer);
                    programTextArea.setBounds(0, 0, 0, 0);
                    programTextAreaPeer.add(programTextArea);
                }
                else
                    programPanel.add(programTextArea);
            }

            warningLabel = new Label("Compiler Warnings:");
            warningLabel.setFont(LABEL_FONT);
            leftPanel.add(
                warningLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 2,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            warningList = new List(8);
            warningList.addFocusListener(focusManager);
            warningList.addItemListener(this);
            warningList.setBackground(new Color(192, 192, 192));
            warningList.setFont(MONOSPACED_FONT);
            warningList.setForeground(new Color(0, 0, 0));
            warningList.setMultipleMode(true);
            leftPanel.add(
                warningList,
                ToyToolkit.createGridBagConstraints(
                    0, 3,
                    1, 1,
                    1.0, 40.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
        }

        rightPanel = new Panel(new GridBagLayout());
        rightPanel.setBackground(getBackground());
        add(
            rightPanel,
            ToyToolkit.createGridBagConstraints(
                1, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5),
                0, 0
                )
            );
        {
            inputLabel = new Label("Input File:");
            inputLabel.setFont(LABEL_FONT);
            rightPanel.add(
                inputLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 0,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            inputPanel = new ToyInputPanel(focusManager);
            rightPanel.add(
                inputPanel,
                ToyToolkit.createGridBagConstraints(
                    0, 1,
                    1, 1,
                    1.0, 10.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            referenceLabel = new Label("X-TOY Instruction Set:");
            referenceLabel.setFont(LABEL_FONT);
            rightPanel.add(
                referenceLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 2,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );


            referenceTextArea = new TextArea(ToyVirtualMachine.REFERENCE_STRING);
            referenceTextArea.addFocusListener(focusManager);
            referenceTextArea.setBackground(new Color(192, 192, 192));
            referenceTextArea.setEditable(false);
            referenceTextArea.setFont(MONOSPACED_FONT);
            referenceTextArea.setForeground(new Color(0, 0, 0));
            rightPanel.add(
                referenceTextArea,
                ToyToolkit.createGridBagConstraints(
                    0, 3,
                    1, 1,
                    1.0, 60.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            converterLabel = new Label("16-bit Base Converter:");
            converterLabel.setFont(LABEL_FONT);
            rightPanel.add(
                converterLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 4,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            converterPanel = new ToyConverterPanel(focusManager);
            rightPanel.add(
                converterPanel,
                ToyToolkit.createGridBagConstraints(
                    0, 5,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // setWorkspace() allows the frame to change the workspace displayed by this
    // panel
    public void setWorkspace(ToyWorkspace newWorkspace)
    {
        //on our first call, do the netscape test
        if (workspace == null)
            programTextArea.testForNetscapeBug();

        workspace = newWorkspace;
        workspace.reset();
        inputPanel.setVirtualMachine(workspace.getInitialVirtualMachine());
        programTextArea.setText(workspace.getText());
        refresh(true);
    }

    ////////////////////////////////////////////////////////////////////////////
    // intercept setBounds() calls
    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        super.setBounds(x, y, width, height);

        doLayout();
    }

    ////////////////////////////////////////////////////////////////////////////
    // refresh() updates the text in the programTextArea with the text of the
    // virtual machine
    public void refresh()
    {
        refresh(false);
    }

    public void refresh(boolean forceRefresh)
    {
        if (!isVisible())
            return;

        workspace.addUndoData(new ToyTextAreaState(programTextArea));

        undoMenuItem.setEnabled(workspace.canUndo());
        redoMenuItem.setEnabled(workspace.canRedo());

        if (forceRefresh || !programTextArea.getText(true).equals(workspace.getText()))
        {
            //pass the program text to the Virtual Machine for parsing
            warningHash = workspace.setText(
                programTextArea.getText(true),
                warningCheckbox.getState()
                );

            //update the label
            if (workspace.getFileName() == null ||
                workspace.getFileName().length() == 0)
                programLabel.setText(workspace.getName() + " Code:");
            else
                programLabel.setText(workspace.getName() + " Code: (" +
                    workspace.getFileName() + ")");

            //dump the warnings into the warninglist
            warningList.removeAll();
            for (int ctr = 0; ctr < warningHash.getCardinality(); ctr++)
                warningList.add(warningHash.getText(ctr), ctr);
            warningList.paint(warningList.getGraphics());
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // testAutocomment() tests each keystroke for autocommenting
    public void testAutocomment(KeyEvent e)
    {
        //if a backspace is typed immediately following an autocomment, undo it
        if (focusManager.getFocusedComponent() == programTextArea &&
            backspaceUndoableAutoComment && e.getKeyChar() == 8)
        {
            workspace.undo().restoreState(programTextArea);
            refresh();
        }

        backspaceUndoableAutoComment = false;

        //get ready to autocomment
        if (focusManager.getFocusedComponent() == programTextArea &&
            autoCommentCheckbox.getState() &&
            (ToyWord.isHexDigit(e.getKeyChar()) ||
             e.getKeyChar() == ' ' ||
             e.getKeyChar() == ':')
            )
        {
            int caretPosition = programTextArea.getCaretPosition();
            int startPosition = caretPosition;
            int endPosition;
            boolean moveCaretToEnd;
            String text = programTextArea.getText(true);

            //locate the first significant character of this line
            startPosition--;
            while (startPosition >= 0 &&
                   text.charAt(startPosition) != '\n')
                startPosition--;

            startPosition++;
            while (startPosition < text.length() &&
                   Character.isWhitespace(text.charAt(startPosition)) &&
                   text.charAt(startPosition) != ' ')
                startPosition++;

            //make sure that our caret was in a position to have just added
            //a character something of the form "##: ####"
            if (caretPosition <= startPosition ||
                caretPosition > startPosition + 8)
                return;

            //make sure that we found something of the form "##: ####"
            if (startPosition + 7 >= text.length() ||
                !ToyWord.isHexDigit(text.charAt(startPosition)) ||
                !ToyWord.isHexDigit(text.charAt(startPosition + 1)) ||
                text.charAt(startPosition + 2) != ':' ||
                text.charAt(startPosition + 3) != ' ' ||
                !ToyWord.isHexDigit(text.charAt(startPosition + 4)) ||
                !ToyWord.isHexDigit(text.charAt(startPosition + 5)) ||
                !ToyWord.isHexDigit(text.charAt(startPosition + 6)) ||
                !ToyWord.isHexDigit(text.charAt(startPosition + 7)))
                return;

            //if our caret was at the end of the "##: ####" then we might be
            //able to move the caret to the end of the comment when we're done
            //(see next code block for additional stipulations)
            moveCaretToEnd = (caretPosition == startPosition + 8);

            //find the end position
            //(if there was significant text between the "##: ####" and
            // the endPosition, then we are not to move the caret to the end
            // of the comment when this is over)
            endPosition = startPosition + 8;
            while (endPosition < text.length() &&
                   endPosition < startPosition + 8 + ToyWorkspace.COMMENT_WIDTH &&
                   text.charAt(endPosition) != '\n')
            {
                if (!Character.isWhitespace(text.charAt(endPosition)))
                    moveCaretToEnd = false;
                endPosition++;
            }

            //ok we just earned ourselves an autocomment!
            String commentString;
            ToyWord value = ToyWord.parseHexidecimal(
                text.substring(
                    startPosition + 4,
                    startPosition + 8
                    )
                );

            if (text.charAt(startPosition) == '0')
                commentString = "   " + value.toString();
            else
                commentString = "   " + value.toPseudoCodeString();

            while (commentString.length() < ToyWorkspace.COMMENT_WIDTH)
                commentString += " ";

            //back up the current state
            workspace.addUndoData(new ToyTextAreaState(programTextArea));

            programTextArea.setText(
                text.substring(0, startPosition + 8) +
                    commentString +
                    text.substring(endPosition),
                false
                );

            if (moveCaretToEnd)
                programTextArea.setCaretPosition(startPosition + 8 + ToyWorkspace.COMMENT_WIDTH);
            else
                programTextArea.setCaretPosition(caretPosition);
            backspaceUndoableAutoComment = true;
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ActionListener
    public void actionPerformed(ActionEvent e)
    {
        if (e.getSource() == undoMenuItem)
        {
            workspace.undo().restoreState(programTextArea);
            refresh();
        }

        if (e.getSource() == redoMenuItem)
        {
            workspace.redo().restoreState(programTextArea);
            refresh();
        }

        if (e.getSource() == insertFunctionCommentMenuItem)
        {
            refresh();

            if (focusManager.getFocusedComponent() == programTextArea)
            {
                int positionFromStart = programTextArea.getSelectionStart();
                int positionFromEnd = programTextArea.getText(true).length() -
                                      programTextArea.getSelectionEnd();

                programTextArea.setText(
                    programTextArea.getText(true).substring(0, programTextArea.getSelectionStart()) +
                    "////////////////////////////////////////////////////////////////////////////////\n" +
                    "// Function \n" +
                    "// Input:                \n" +
                    "// Return address:       \n" +
                    "// Output:               \n" +
                    "// Temporary variables:  " +
                    programTextArea.getText(true).substring(programTextArea.getSelectionEnd())
                    );

                programTextArea.requestFocus();
                programTextArea.select(
                    positionFromStart,
                    programTextArea.getText(true).length() - positionFromEnd
                    );
            }
        }


        if (e.getSource() == appendCommentMenuItem)
        {
            refresh();

            if (workspace.hasFatalError())
            {
                ToyPopup.askUser(
                    this,
                    "Error",
                    "Your workspace has fatal errors\n" +
                    "\tVisual X-TOY cannot update the comments of a program with " +
                    "fatal errors.  Please fix these errors and try again.",
                    "OK"
                    );
                return;
            }

            int line = 0, ctr;

            for (ctr = 0; ctr < programTextArea.getCaretPosition(); ctr++)
                if (programTextArea.getText(true).charAt(ctr) == '\n')
                    line++;

            programTextArea.setText(
                workspace.getModifiedText()
                );

            refresh();

            if (focusManager.getFocusedComponent() == programTextArea)
            {
                String text = programTextArea.getText(true);
                for (ctr = 0; line > -1 && ctr < text.length(); ctr++)
                    if (text.charAt(ctr) == '\n')
                        line--;

                if (ctr > 0 && ctr < text.length())
                    ctr--;

                programTextArea.requestFocus();
                programTextArea.setCaretPosition(ctr);
            }
        }



        if (e.getSource() == reformatCodeMenuItem)
        {
            refresh();

            if (workspace.hasFatalError())
            {
                ToyPopup.askUser(
                    this,
                    "Error",
                    "Your workspace has fatal errors\n" +
                    "\tVisual X-TOY cannot reformat a program with fatal errors.  " +
                    "Please fix these errors and try again.",
                    "OK"
                    );
                return;
            }

            if (ToyPopup.askUser(
                    this,
                    "Reformat Code Confirmation",
                    "Are you sure you wish to completely reformat your code?\n" +
                    "\tAll of your comments will be erased and replaced with Visual " +
                    "X-TOY's comments!",
                    "Yes",
                    "No"
                    )
                )
            {
                int barIndex = workspace.getText().indexOf("\n" + ToyWorkspace.HEADER_BAR + "\n");
                if (barIndex >= 0)
                    programTextArea.setText(
                        workspace.getModifiedText(
                            workspace.getText().substring(0, barIndex) + "\n" +
                            ToyWorkspace.HEADER_BAR + "\n"
                            )
                        );
                else
                    programTextArea.setText(
                        workspace.getModifiedText(
                            "Program " + workspace.getName() + "\n" +
                            "Input:    \n" +
                            "Output:   \n" +
                            "Remarks:  \n" +
                            ToyWorkspace.HEADER_BAR + "\n" +
                            "\n"
                            )
                        );
                refresh();
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ItemListener
    public void itemStateChanged(ItemEvent e)
    {
        //if the warningCheckbox was changed, then refresh the warningList
        if (e.getSource() == warningCheckbox && isVisible())
            refresh(true);

        //if the highlightingCheckbox was changed, make the correct textarea
        //visible
        if (e.getSource() == highlightingCheckbox)
        {
            if (highlightingCheckbox.getState())
            {
                programPanel.remove(programTextArea);
                programPanel.add(programTextAreaPeer);
                programTextArea.setBounds(0, 0, 0, 0);
                programTextAreaPeer.add(programTextArea);
                programPanel.doLayout();
            }
            else {
                programPanel.remove(programTextAreaPeer);
                programTextAreaPeer.remove(programTextArea);
                programPanel.add(programTextArea);
                programPanel.doLayout();
            }
        }

        //if a line in the warningList was selected, then highlight the
        //corresponding line in the program text
        if (e.getSource() == warningList)
        {
            String programText;
            int begin, end, line, lineCtr = 0;
            int selectedIndex = ((Integer)e.getItem()).intValue();


            //retrieve the line to zoom to
            line = warningHash.getCitation(selectedIndex);

            //grab a copy of the program text and look for the character index of
            //the beginning of that line
            programText = programTextArea.getText(true);
            begin = 0;
            while (begin < programText.length() && lineCtr < line - 1)
            {
                if (programText.charAt(begin) == '\n')
                    lineCtr++;
                begin++;
            }

            //select that entire line
            end = begin + 1;
            while (end < programText.length() && programText.charAt(end) != '\n')
                end++;

            //And select them
            programTextArea.requestFocus();
            programTextArea.setCaretPosition(begin);
            programTextArea.select(begin, end);

            //highlight the entire group
            int selected[] = warningList.getSelectedIndexes();
            for (int ctr = 0; ctr < selected.length; ctr++)
                warningList.deselect(selected[ctr]);

            begin = selectedIndex;
            end = selectedIndex;

            while (begin - 1 >= 0 &&
                   warningHash.getIndex(begin - 1) == warningHash.getIndex(selectedIndex))
                begin--;
            while (end + 1 < warningHash.getCardinality() &&
                   warningHash.getIndex(end + 1) == warningHash.getIndex(selectedIndex))
                end++;

            for (int ctr = end; ctr >= begin; ctr--)
                warningList.select(ctr);

            warningList.requestFocus();
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement KeyListener
    public void keyPressed(KeyEvent e)
    {
        testAutocomment(e);
    }

    public void keyReleased(KeyEvent e)
    {
        testAutocomment(e);
    }

    public void keyTyped(KeyEvent e)
    {
        testAutocomment(e);
    }
}