package toy.dialog;

import java.awt.*;
import java.awt.event.*;

import toy.image.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyImageToggleButton extends Component implements ItemSelectable
{
    private int disabledImageIndex,
                offImageIndex, hoverOffImageIndex, activeOffImageIndex,
                onImageIndex, hoverOnImageIndex, activeOnImageIndex;
    private ToyImageManager imageManager;

    private ItemListener itemListener;

    private boolean mouseInside;
    private boolean mouseDown;
    private boolean hasFocus;

    private boolean state;


    public ToyImageToggleButton(ToyImageManager newImageManager,
                                int newDisabledImageIndex, int newOffImageIndex,
                                int newHoverOffImageIndex, int newActiveOffImageIndex,
                                int newOnImageIndex, int newHoverOnImageIndex,
                                int newActiveOnImageIndex)
    {
        imageManager = newImageManager;

        disabledImageIndex = newDisabledImageIndex;
        offImageIndex = newOffImageIndex;
        hoverOffImageIndex = newHoverOffImageIndex;
        activeOffImageIndex = newActiveOffImageIndex;
        onImageIndex = newOnImageIndex;
        hoverOnImageIndex = newHoverOnImageIndex;
        activeOnImageIndex = newActiveOnImageIndex;

        enableEvents(AWTEvent.FOCUS_EVENT_MASK);
        enableEvents(AWTEvent.KEY_EVENT_MASK);
        enableEvents(AWTEvent.MOUSE_EVENT_MASK);
        enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);
    }



    ////////////////////////////////////////////////////////////////////////////
    // Describe our new Component

    public boolean isFocusTraversable()
    {
        return isEnabled();
    }

    public boolean isDoubleBuffered()
    {
        return false;
    }

    public boolean isLightweight()
    {
        return true;
    }

    public Dimension getPreferredSize()
    {
        return new Dimension(
            imageManager.getImage(disabledImageIndex).getWidth(null),
            imageManager.getImage(disabledImageIndex).getHeight(null)
            );
    }

    public Dimension getMinimumSize()
    {
        return new Dimension(
            imageManager.getImage(disabledImageIndex).getWidth(null),
            imageManager.getImage(disabledImageIndex).getHeight(null)
            );
    }

    public Dimension getMaximumSize()
    {
        return new Dimension(
            imageManager.getImage(disabledImageIndex).getWidth(null),
            imageManager.getImage(disabledImageIndex).getHeight(null)
            );
    }

    ////////////////////////////////////////////////////////////////////////////
    // The get and set functions

    public boolean hasFocus()
    {
        return hasFocus;
    }

    public void setEnabled(boolean b)
    {
        super.setEnabled(b);
        repaint();
    }

    public void setVisible(boolean b)
    {
        if (b == isVisible())
            return;

        super.setVisible(b);

        if (getParent() != null)
        {
            if (b)
                repaint();
            else
                getParent().repaint(
                    getBounds().x,
                    getBounds().y,
                    getBounds().width,
                    getBounds().height
                    );
        }
    }

    public void setState(boolean newState)
    {
        //changes in state from the outside don't generate events
        setState(newState, false);
    }

    private void setState(boolean newState, boolean dispatchEvent)
    {
        if (state == newState)
            return;

        state = newState;

        if (itemListener != null && dispatchEvent)
        {
            if (state)
                itemListener.itemStateChanged(
                    new ItemEvent(
                        this,
                        ItemEvent.ITEM_STATE_CHANGED,
                        this,
                        ItemEvent.SELECTED
                        )
                    );
            else
                itemListener.itemStateChanged(
                    new ItemEvent(
                        this,
                        ItemEvent.ITEM_STATE_CHANGED,
                        this,
                        ItemEvent.DESELECTED
                        )
                    );
        }

        repaint();
    }

    public boolean getState()
    {
        return state;
    }

    public void setImageIndices(int newDisabledImageIndex, int newOffImageIndex,
                                int newHoverOffImageIndex, int newActiveOffImageIndex,
                                int newOnImageIndex, int newHoverOnImageIndex,
                                int newActiveOnImageIndex)
    {
        disabledImageIndex = newDisabledImageIndex;
        offImageIndex = newOffImageIndex;
        hoverOffImageIndex = newHoverOffImageIndex;
        activeOffImageIndex = newActiveOffImageIndex;
        onImageIndex = newOnImageIndex;
        hoverOnImageIndex = newHoverOnImageIndex;
        activeOnImageIndex = newActiveOnImageIndex;

        repaint();
    }

    public int getDisabledImageIndex()
    {
        return disabledImageIndex;
    }

    public int getOffImageIndex()
    {
        return offImageIndex;
    }

    public int getHoverOffImageIndex()
    {
        return hoverOffImageIndex;
    }

    public int getActiveOffImageIndex()
    {
        return activeOffImageIndex;
    }

    public int getOnImageIndex()
    {
        return onImageIndex;
    }

    public int getHoverOnImageIndex()
    {
        return hoverOnImageIndex;
    }

    public int getActiveOnImageIndex()
    {
        return activeOnImageIndex;
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ItemSelectable

    public void addItemListener(ItemListener newItemListener)
    {
        itemListener = AWTEventMulticaster.add(itemListener, newItemListener);
    }


    public void removeItemListener(ItemListener newItemListener)
    {
       itemListener = AWTEventMulticaster.remove(itemListener, newItemListener);
    }

    public Object[] getSelectedObjects()
    {
        if (state)
        {
            Object answer[] = new Object[1];
            answer[0] = this;
            return answer;
        }
        else
            return null;
    }


    ////////////////////////////////////////////////////////////////////////////
    // Overriding the painting functions
    public void update(Graphics g)
    {
        paint(g);
    }

    public void paint(Graphics g)
    {
        if (getParent() == null || !isVisible() || g == null)
            return;

        if (!isEnabled())
            g.drawImage(
                imageManager.getImage(disabledImageIndex),
                0,
                0,
                this
                );
        else {
            if (state)
            {
                if (mouseInside && mouseDown)
                    g.drawImage(
                        imageManager.getImage(activeOnImageIndex),
                        0,
                        0,
                        this
                        );
                else {
                    if (mouseInside || hasFocus)
                        g.drawImage(
                            imageManager.getImage(hoverOnImageIndex),
                            0,
                            0,
                            this
                            );
                    else
                        g.drawImage(
                            imageManager.getImage(onImageIndex),
                            0,
                            0,
                            this
                            );
                }
            }
            else {
                if (mouseInside && mouseDown)
                    g.drawImage(
                        imageManager.getImage(activeOffImageIndex),
                        0,
                        0,
                        this
                        );
                else {
                    if (mouseInside || hasFocus)
                        g.drawImage(
                            imageManager.getImage(hoverOffImageIndex),
                            0,
                            0,
                            this
                            );
                    else
                        g.drawImage(
                            imageManager.getImage(offImageIndex),
                            0,
                            0,
                            this
                            );
                }
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept mouse motion events

    public void processMouseMotionEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        switch (e.getID())
        {
            case MouseEvent.MOUSE_MOVED:
                break;

            case MouseEvent.MOUSE_DRAGGED:
                if (e.getX() < 0 || e.getY() < 0 || e.getX() > getBounds().width ||
                    e.getY() > getBounds().height)
                    mouseInside = false;
                else
                    mouseInside = true;
                repaint();
                break;
        }

        super.processMouseMotionEvent(e);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept mouse events

    public void processMouseEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        switch (e.getID())
        {
            case MouseEvent.MOUSE_CLICKED:
                //MOUSE_RELEASED is much easier to satisfy for a "click"
                break;

            case MouseEvent.MOUSE_PRESSED:
                requestFocus();
                mouseDown = true;
                repaint();
                break;

            case MouseEvent.MOUSE_RELEASED:
                mouseDown = false;
                if (mouseInside)
                    setState(!state, true);
                repaint();
                break;

            case MouseEvent.MOUSE_ENTERED:
                mouseInside = true;
                repaint();
                break;

            case MouseEvent.MOUSE_EXITED:
                mouseInside = false;
                repaint();
                break;
        }

        super.processMouseEvent(e);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept focus events

    public void processFocusEvent(FocusEvent e)
    {
        switch (e.getID())
        {
            case FocusEvent.FOCUS_GAINED:
                hasFocus = true;
                repaint();
                break;

            case FocusEvent.FOCUS_LOST:
                hasFocus = false;
                repaint();
                break;
        }

        super.processFocusEvent(e);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept key events

    public void processKeyEvent(KeyEvent e)
    {
        if (!isEnabled())
            return;

        if (e.getID() == KeyEvent.KEY_PRESSED)
        {
            switch(e.getKeyChar())
            {
                case ' ':
                case '\n':
                    setState(!state, true);
                    repaint();
                    break;
            }
        }

        super.processKeyEvent(e);
    }
}