package toy.dialog;

import java.awt.*;
import java.awt.event.*;
import toy.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyInputPanel extends ToyRefreshablePanel
                           implements ActionListener, KeyListener
{
    ////////////////////////////////////////////////////////////////////////////
    // constants
    public static final Font MONOSPACED_FONT = new Font("Monospaced", Font.PLAIN, 12);
    public static final Font SANSSERIF_FONT = new Font("SansSerif", Font.PLAIN, 12);
    public static final Font LABEL_FONT = new Font("SansSerif", Font.BOLD, 12);

    ////////////////////////////////////////////////////////////////////////////
    // variables
    private ToyVirtualMachine virtualMachine;

    private ToyFocusManager focusManager;

    private Panel leftPanel;
        private TextField inputTextField;
        private List inputList;

    private Panel rightPanel;
        private Button addButton;
        private Button editButton;

    ////////////////////////////////////////////////////////////////////////////
    // the constructor
    public ToyInputPanel(ToyFocusManager newFocusManager)
    {
        super(new GridBagLayout());

        focusManager = newFocusManager;

        setBackground(new Color(192, 192, 192));

        leftPanel = new Panel(new GridBagLayout());
        add(
            leftPanel,
            ToyToolkit.createGridBagConstraints(
                0, 0,
                1, 1,
                3.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5),
                0, 0
                )
            );
        {
            inputTextField = new TextField(5);
            inputTextField.addFocusListener(focusManager);
            inputTextField.addKeyListener(this);
            inputTextField.setBackground(new Color(255, 255, 255));
            inputTextField.setFont(MONOSPACED_FONT);
            inputTextField.setForeground(new Color(0, 0, 0));
            leftPanel.add(
                inputTextField,
                ToyToolkit.createGridBagConstraints(
                    0, 0,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.NORTH,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 5, 0),
                    0, 0
                    )
                );

            inputList = new List(5);
            inputList.addFocusListener(focusManager);
            inputList.setBackground(new Color(255, 255, 255));
            inputList.setFont(MONOSPACED_FONT);
            inputList.setForeground(new Color(0, 0, 0));
            leftPanel.add(
                inputList,
                ToyToolkit.createGridBagConstraints(
                    0, 1,
                    1, 1,
                    1.0, 100.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
        }


        rightPanel = new Panel(new GridLayout(3, 1, 0, 5));
        add(
            rightPanel,
            ToyToolkit.createGridBagConstraints(
                1, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.NORTH,
                GridBagConstraints.HORIZONTAL,
                new Insets(5, 0, 5, 5),
                0, 0
                )
            );


        {
            addButton = new Button("Add");
            addButton.addActionListener(this);
            addButton.addFocusListener(focusManager);
            addButton.setBackground(new Color(192, 192, 192));
            addButton.setForeground(new Color(0, 0, 0));
            rightPanel.add(addButton);

            editButton = new Button("Edit");
            editButton.addActionListener(this);
            editButton.addFocusListener(focusManager);
            editButton.setBackground(new Color(192, 192, 192));
            editButton.setForeground(new Color(0, 0, 0));
            rightPanel.add(editButton);

            rightPanel.add(new ToySpacer(80, 0));
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // addRequest() adds the contents of the inputTextField to stdin
    public void addRequest()
    {
        //change stdin based on the contents of the Text Area
        String text = inputTextField.getText();
        ToyWord stdin[] = virtualMachine.getStdin();

        inputTextField.setText("");

        int beginCtr = 0, endCtr = 0;

        while (endCtr < text.length())
        {
            beginCtr = endCtr;

            //advance beginCtr to the next hex digit
            while (beginCtr < text.length() &&
                   !ToyWord.isHexDigit(text.charAt(beginCtr)))
                beginCtr++;

            //advance endCtr to the next non-hex digit or 4 characters,
            //whichever is less
            endCtr = beginCtr + 1;
            while (endCtr < text.length() &&
                   endCtr < beginCtr + 4 &&
                   ToyWord.isHexDigit(text.charAt(endCtr)))
                endCtr++;

            if (beginCtr < text.length())
            {
                char charArray[] = new char[4];
                ToyWord oldStdin[] = stdin;
                stdin = new ToyWord[oldStdin.length + 1];

                for (int ctr = 0; ctr < oldStdin.length; ctr++)
                    stdin[ctr] = oldStdin[ctr];

                for (int ctr = 0; ctr < 4; ctr++)
                {
                    if (endCtr - 4 + ctr < beginCtr)
                        charArray[ctr] = '0';
                    else
                        charArray[ctr] = text.charAt(endCtr - 4 + ctr);
                }

                stdin[oldStdin.length] = ToyWord.parseHexidecimal(
                    charArray[0],
                    charArray[1],
                    charArray[2],
                    charArray[3]
                    );
            }
        }

        virtualMachine.setStdin(stdin);

        refresh();
    }

    ////////////////////////////////////////////////////////////////////////////
    // setVirtualMachine() allows the frame to change the workspace displayed by
    // this panel
    public void setVirtualMachine(ToyVirtualMachine newVirtualMachine)
    {
        virtualMachine = newVirtualMachine;

        inputTextField.setText("");

        refresh();
    }

    ////////////////////////////////////////////////////////////////////////////
    // intercept setBounds() calls
    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        super.setBounds(x, y, width, height);

        doLayout();
    }

    ////////////////////////////////////////////////////////////////////////////
    // refresh() is the method called when the list needs to be updated
    public void refresh()
    {
        inputList.setVisible(false);

        inputList.removeAll();

        ToyWord stdin[] = virtualMachine.getStdin();

        for (int ctr = 0; ctr < virtualMachine.getStdinCtr(); ctr++)
            inputList.add("*" + stdin[ctr].toHexString());

        for (int ctr = virtualMachine.getStdinCtr(); ctr < stdin.length; ctr++)
            inputList.add(" " + stdin[ctr].toHexString());

        inputList.add("(Prompt)");

        inputList.select(virtualMachine.getStdinCtr());

        inputList.setVisible(true);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ActionListener
    public void actionPerformed(ActionEvent e)
    {
        int selectedIndex = inputList.getSelectedIndex();

        if (e.getSource() == addButton)
            addRequest();

        if (e.getSource() == editButton)
        {
            ToyPopup.editInput(
                this,
                virtualMachine
                );
            refresh();
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement KeyListener
    public void keyPressed(KeyEvent e)
    {
        if (e.getKeyChar() == '\n')
            addRequest();
    }

    public void keyReleased(KeyEvent e)
    {
        if (e.getKeyChar() == '\n')
            addRequest();
    }

    public void keyTyped(KeyEvent e)
    {
        if (e.getKeyChar() == '\n')
            addRequest();
    }
}