package toy.dialog;

import java.awt.*;
import java.awt.event.*;

import toy.*;
import toy.image.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToySimPanel extends ToyRefreshablePanel
                         implements ActionListener, AdjustmentListener,
                                    ItemListener
{
    public static final int SIM_UNIT_WIDTH = 104;
    public static final int SIM_UNIT_HEIGHT = 67;

    public static final int ADDR_CARDINALITY = 8;
    public static final int OUTPUT_CARDINALITY = 4;
    public static final int DATA_CARDINALITY = 16;

    public static final int VIEW_ROWS    = 16;
    public static final int VIEW_COLUMNS = 2;

    private ToyFocusManager focusManager;
    private ToyImageManager imageManager;

    private ToyWorkspace workspace;

    private CheckboxMenuItem closeExecutionCheckbox;

    private MenuItem stepMenuItem;
    private MenuItem runMenuItem;
    private MenuItem resetMenuItem;

    //Dialog items
    private Panel machinePanel;
        private ToyImageButton machineLoadButton;
        private ToyImageButton machineLookButton;

        private ToyImageButton machineStepButton;
        private ToyImageButton machineRunButton;

        private ToyImageButton machineResetButton;

        private ToyImageLabel machineAddrLabel;
        private ToyImageLabel machineAddrLight[];
        private ToyImageToggleButton machineAddrSwitch[];

        private ToyImageLabel machineOutputLabel;
        private ToyImageLabel machineOutputLCD[];

        private ToyImageLabel machineDataLabel;
        private ToyImageLabel machineDataLight[];
        private ToyImageToggleButton machineDataSwitch[];

    private Panel viewPanel;
        private Choice viewChoice;
        private Label viewLabel[];
        private Scrollbar viewScrollbar;

    private Panel registerPanel;
        private Label registerLabel[];
        private Label registerContent[];

        private Label registerCurrentLabel;
        private Label registerCurrentHex;
        private Label registerCurrentPseudocode;

        private Label registerCommandLabel;
        private Label registerCommandHex;
        private Label registerCommandPseudocode;

    ////////////////////////////////////////////////////////////////////////////
    // the constructor
    public ToySimPanel(ToyFocusManager newFocusManager,
                       ToyImageManager newImageManager,
                       CheckboxMenuItem newCloseExecutionCheckbox,
                       MenuItem newStepMenuItem,
                       MenuItem newRunMenuItem,
                       MenuItem newResetMenuItem
                       )
    {
        super(null);

        focusManager = newFocusManager;
        imageManager = newImageManager;

        closeExecutionCheckbox = newCloseExecutionCheckbox;

        stepMenuItem = newStepMenuItem;
            stepMenuItem.addActionListener(this);
        runMenuItem = newRunMenuItem;
            runMenuItem.addActionListener(this);
        resetMenuItem = newResetMenuItem;
            resetMenuItem.addActionListener(this);

        setBackground(new Color(80, 100, 120));

        machinePanel = new Panel(null);
        machinePanel.setBackground(new Color (80, 100, 120));
        add(machinePanel);

            machineLoadButton = new ToyImageButton(
                imageManager,
                ToyImageManager.FRAME_SIM_LOAD_OFF,
                ToyImageManager.FRAME_SIM_LOAD_ON,
                ToyImageManager.FRAME_SIM_LOAD_ON,
                ToyImageManager.FRAME_SIM_LOAD_OFF
                );
            machineLoadButton.addActionListener(this);
            machinePanel.add(machineLoadButton);

            machineLookButton = new ToyImageButton(
                imageManager,
                ToyImageManager.FRAME_SIM_LOOK_OFF,
                ToyImageManager.FRAME_SIM_LOOK_ON,
                ToyImageManager.FRAME_SIM_LOOK_ON,
                ToyImageManager.FRAME_SIM_LOOK_OFF
                );
            machineLookButton.addActionListener(this);
            machinePanel.add(machineLookButton);

            machineStepButton = new ToyImageButton(
                imageManager,
                ToyImageManager.FRAME_SIM_STEP_OFF,
                ToyImageManager.FRAME_SIM_STEP_ON,
                ToyImageManager.FRAME_SIM_STEP_ON,
                ToyImageManager.FRAME_SIM_STEP_OFF
                );
            machineStepButton.addActionListener(this);
            machinePanel.add(machineStepButton);

            machineRunButton = new ToyImageButton(
                imageManager,
                ToyImageManager.FRAME_SIM_RUN_OFF,
                ToyImageManager.FRAME_SIM_RUN_ON,
                ToyImageManager.FRAME_SIM_RUN_ON,
                ToyImageManager.FRAME_SIM_RUN_OFF
                );
            machineRunButton.addActionListener(this);
            machinePanel.add(machineRunButton);

            machineResetButton = new ToyImageButton(
                imageManager,
                ToyImageManager.FRAME_SIM_RESET_OFF,
                ToyImageManager.FRAME_SIM_RESET_ON,
                ToyImageManager.FRAME_SIM_RESET_ON,
                ToyImageManager.FRAME_SIM_RESET_OFF
                );
            machineResetButton.addActionListener(this);
            machinePanel.add(machineResetButton);

            machineAddrLabel = new ToyImageLabel(
                imageManager,
                ToyImageManager.FRAME_SIM_ADDR
                );
            machinePanel.add(machineAddrLabel);

            machineAddrLight = new ToyImageLabel[ADDR_CARDINALITY];
            machineAddrSwitch = new ToyImageToggleButton[ADDR_CARDINALITY];
            for (int ctr = 0; ctr < ADDR_CARDINALITY; ctr++)
            {
                machineAddrLight[ctr] = new ToyImageLabel(
                    imageManager,
                    ToyImageManager.FRAME_SIM_LIGHT_OFF
                    );
                machinePanel.add(machineAddrLight[ctr]);
                machineAddrSwitch[ctr] = new ToyImageToggleButton(
                    imageManager,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF,
                    ToyImageManager.FRAME_SIM_SWITCH_ON,
                    ToyImageManager.FRAME_SIM_SWITCH_ON,
                    ToyImageManager.FRAME_SIM_SWITCH_ON,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF
                    );
                machineAddrSwitch[ctr].addItemListener(this);
                machinePanel.add(machineAddrSwitch[ctr]);
            }

            machineOutputLabel = new ToyImageLabel(
                imageManager,
                ToyImageManager.FRAME_SIM_OUTPUT
                );
            machinePanel.add(machineOutputLabel);

            machineOutputLCD = new ToyImageLabel[OUTPUT_CARDINALITY];
            for (int ctr = 0; ctr < OUTPUT_CARDINALITY; ctr++)
            {
                machineOutputLCD[ctr] = new ToyImageLabel(
                    imageManager,
                    ToyImageManager.FRAME_SIM_DIGIT_BLANK
                    );
                machinePanel.add(machineOutputLCD[ctr]);
            }

            machineDataLabel = new ToyImageLabel(
                imageManager,
                ToyImageManager.FRAME_SIM_DATA
                );
            machinePanel.add(machineDataLabel);

            machineDataLight = new ToyImageLabel[DATA_CARDINALITY];
            machineDataSwitch = new ToyImageToggleButton[DATA_CARDINALITY];
            for (int ctr = 0; ctr < DATA_CARDINALITY; ctr++)
            {
                machineDataLight[ctr] = new ToyImageLabel(
                    imageManager,
                    ToyImageManager.FRAME_SIM_LIGHT_OFF
                    );
                machinePanel.add(machineDataLight[ctr]);
                machineDataSwitch[ctr] = new ToyImageToggleButton(
                    imageManager,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF,
                    ToyImageManager.FRAME_SIM_SWITCH_ON,
                    ToyImageManager.FRAME_SIM_SWITCH_ON,
                    ToyImageManager.FRAME_SIM_SWITCH_ON,
                    ToyImageManager.FRAME_SIM_SWITCH_OFF
                    );
                machineDataSwitch[ctr].addItemListener(this);
                machinePanel.add(machineDataSwitch[ctr]);
            }

        viewPanel = new Panel(null);
        viewPanel.setBackground(new Color (192, 192, 192));
        add(viewPanel);

            viewChoice = new Choice();
            viewChoice.add("Memory");
            viewChoice.add("Stdin");
            viewChoice.add("Stdout");
            viewChoice.select(0);
            viewChoice.setForeground(new Color(0, 0, 0));
            viewChoice.addItemListener(this);
            viewPanel.add(viewChoice);

            viewLabel = new Label[VIEW_COLUMNS * VIEW_ROWS];
            for (int ctr = 0; ctr < VIEW_COLUMNS * VIEW_ROWS; ctr++)
            {
                viewLabel[ctr] = new Label("00:0000", Label.CENTER);
                viewLabel[ctr].setForeground(new Color(0, 0, 0));
                viewPanel.add(viewLabel[ctr]);
            }

            viewScrollbar = new Scrollbar(
                Scrollbar.HORIZONTAL,
                0,
                VIEW_COLUMNS,
                0,
                (int)Math.ceil(1.0 * ToyVirtualMachine.MEM_CARDINALITY / VIEW_ROWS)
                );
            viewScrollbar.addAdjustmentListener(this);
            viewPanel.add(viewScrollbar);


        registerPanel = new Panel(null);
        registerPanel.setBackground(new Color (192, 192, 192));
        add(registerPanel);

            registerLabel = new Label[ToyVirtualMachine.REGISTER_CARDINALITY];
            registerContent = new Label[ToyVirtualMachine.REGISTER_CARDINALITY];
            for (short ctr = 0; ctr < ToyVirtualMachine.REGISTER_CARDINALITY; ctr++)
            {
                registerLabel[ctr] = new Label("R[" +
                    ToyWord.convertToHexDigit(ctr) + "]");
                registerLabel[ctr].setForeground(new Color(0, 0, 0));
                registerPanel.add(registerLabel[ctr]);

                registerContent[ctr] = new Label("0000");
                registerContent[ctr].setForeground(new Color(0, 0, 0));
                registerPanel.add(registerContent[ctr]);
            }

            registerCurrentLabel = new Label("Current Instruction:");
            registerCurrentLabel.setForeground(new Color(0, 0, 0));
            registerPanel.add(registerCurrentLabel);

            registerCurrentHex = new Label("00:0000");
            registerCurrentHex.setForeground(new Color(0, 0, 0));
            registerPanel.add(registerCurrentHex);

            registerCurrentPseudocode = new Label("halt");
            registerCurrentPseudocode.setForeground(new Color(0, 0, 0));
            registerPanel.add(registerCurrentPseudocode);

            registerCommandLabel = new Label("Current Command:");
            registerCommandLabel.setForeground(new Color(0, 0, 0));
            registerPanel.add(registerCommandLabel);

            registerCommandHex = new Label("00:0000");
            registerCommandHex.setForeground(new Color(0, 0, 0));
            registerPanel.add(registerCommandHex);

            registerCommandPseudocode = new Label("halt");
            registerCommandPseudocode.setForeground(new Color(0, 0, 0));
            registerPanel.add(registerCommandPseudocode);
    }


    ////////////////////////////////////////////////////////////////////////////
    // intercept setBounds() calls
    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        boolean wasVisible = isVisible();
        setVisible(false);

        if (wasVisible)
            getParent().repaint();

        int simUnit = Math.min(width / SIM_UNIT_WIDTH, height / SIM_UNIT_HEIGHT);
        Font sansSerif = new Font("SansSerif", Font.BOLD, 2 * simUnit);
        Font monospaced = new Font("Monospaced", Font.BOLD, 2 * simUnit);

        imageManager.setSimUnit(simUnit);

        machinePanel.setBounds(0, 0, 76 * simUnit, 36 * simUnit);

            machineLoadButton.setBounds (0,            0, 8 * simUnit, 3 * simUnit);
            machineLookButton.setBounds (12 * simUnit, 0, 8 * simUnit, 3 * simUnit);
            machineStepButton.setBounds (24 * simUnit, 0, 8 * simUnit, 3 * simUnit);
            machineRunButton.setBounds  (36 * simUnit, 0, 8 * simUnit, 3 * simUnit);
            machineResetButton.setBounds(68 * simUnit, 0, 8 * simUnit, 3 * simUnit);

            machineAddrLabel.setBounds(0, 4 * simUnit, 8 * simUnit, 2 * simUnit);
            for (int ctr = 0; ctr < ADDR_CARDINALITY; ctr++)
            {
                machineAddrLight[ctr].setBounds(
                    (ctr + ctr / 4) * 4 * simUnit,
                    7 * simUnit,
                    4 * simUnit,
                    4 * simUnit
                    );
                machineAddrSwitch[ctr].setBounds(
                    (ctr + ctr / 4) * 4 * simUnit,
                    11 * simUnit,
                    4 * simUnit,
                    8 * simUnit
                    );
            }

            machineOutputLabel.setBounds(
                44 * simUnit,
                4 * simUnit,
                9 * simUnit,
                2 * simUnit
                );
            for (int ctr = 0; ctr < OUTPUT_CARDINALITY; ctr++)
                machineOutputLCD[ctr].setBounds(
                    (44 + 8 * ctr) * simUnit,
                    7 * simUnit,
                    8 * simUnit,
                    12 * simUnit
                    );

            machineDataLabel.setBounds(0, 21 * simUnit, 8 * simUnit, 2 * simUnit);
            for (int ctr = 0; ctr < DATA_CARDINALITY; ctr++)
            {
                machineDataLight[ctr].setBounds(
                    (ctr + ctr / 4) * 4 * simUnit,
                    24 * simUnit,
                    4 * simUnit,
                    4 * simUnit
                    );
                machineDataSwitch[ctr].setBounds(
                    (ctr + ctr / 4) * 4 * simUnit,
                    28 * simUnit,
                    4 * simUnit,
                    8 * simUnit
                    );
            }

        viewPanel.setBounds(78 * simUnit, 0, 26 * simUnit, 67 * simUnit);

            viewChoice.setBounds(
                2 * simUnit,
                2 * simUnit,
                22 * simUnit,
                3 * simUnit
                );
            viewChoice.setFont(sansSerif);

            for (int ctr = 0; ctr < VIEW_COLUMNS * VIEW_ROWS; ctr++)
            {
                viewLabel[ctr].setBounds(
                    (2 + 11 * (ctr / VIEW_ROWS)) * simUnit,
                    (6 + 3 * (ctr % VIEW_ROWS + (ctr % VIEW_ROWS) / 4)) * simUnit,
                    11 * simUnit,
                    3 * simUnit
                    );
                viewLabel[ctr].setFont(monospaced);
            }

            viewScrollbar.setBounds(
                0,
                65 * simUnit,
                26 * simUnit,
                2 * simUnit
                );

        registerPanel.setBounds(0, 37 * simUnit, 76 * simUnit, 30 * simUnit);

            for (short ctr = 0; ctr < ToyVirtualMachine.REGISTER_CARDINALITY; ctr++)
            {
                registerLabel[ctr].setBounds(
                    (2 + 9 * (ctr % 8)) * simUnit,
                    (2 + 6 * (ctr / 8)) * simUnit,
                    6 * simUnit,
                    2 * simUnit
                    );
                registerLabel[ctr].setFont(monospaced);

                registerContent[ctr].setBounds(
                    (2 + 9 * (ctr % 8)) * simUnit,
                    (4 + 6 * (ctr / 8)) * simUnit,
                    6 * simUnit,
                    2 * simUnit
                    );
                registerContent[ctr].setFont(monospaced);
            }

            registerCurrentLabel.setBounds(
                2 * simUnit,
                14 * simUnit,
                54 * simUnit,
                2 * simUnit
                );
            registerCurrentLabel.setFont(sansSerif);

            registerCurrentHex.setBounds(
                2 * simUnit,
                16 * simUnit,
                54 * simUnit,
                2 * simUnit
                );
            registerCurrentHex.setFont(monospaced);

            registerCurrentPseudocode.setBounds(
                2 * simUnit,
                18 * simUnit,
                54 * simUnit,
                2 * simUnit
                );
            registerCurrentPseudocode.setFont(monospaced);

            registerCommandLabel.setBounds(
                2 * simUnit,
                22 * simUnit,
                54 * simUnit,
                2 * simUnit
                );
            registerCommandLabel.setFont(sansSerif);

            registerCommandHex.setBounds(
                2 * simUnit,
                24 * simUnit,
                54 * simUnit,
                2 * simUnit
                );
            registerCommandHex.setFont(monospaced);

            registerCommandPseudocode.setBounds(
                2 * simUnit,
                26 * simUnit,
                54 * simUnit,
                2 * simUnit
                );
            registerCommandPseudocode.setFont(monospaced);

        super.setBounds(
            x + (width - simUnit * SIM_UNIT_WIDTH) / 2,
            y + (height - simUnit * SIM_UNIT_HEIGHT) / 2,
            simUnit * SIM_UNIT_WIDTH,
            simUnit * SIM_UNIT_HEIGHT
            );

        setVisible(wasVisible);
        refresh();
    }

    ////////////////////////////////////////////////////////////////////////////
    // change the current workspace
    public void setWorkspace(ToyWorkspace newWorkspace)
    {
        workspace = newWorkspace;
        workspace.reset();
        refresh();
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ToyRefreshable
    public void refresh()
    {
        refresh(false);
    }

    public void refresh(boolean supressScrollFix)
    {
        if (workspace != null)
        {
            stepMenuItem.setEnabled(!workspace.getVirtualMachine().isDone());
            runMenuItem.setEnabled(!workspace.getVirtualMachine().isDone());
            machineStepButton.setEnabled(!workspace.getVirtualMachine().isDone());
            machineRunButton.setEnabled(!workspace.getVirtualMachine().isDone());

            String bitString = ToyWord.toBinaryString(
                workspace.getVirtualMachine().getProgramCtr()
                );

            ToyWord stdin[] = workspace.getVirtualMachine().getStdin();
            ToyWord stdout[] = workspace.getVirtualMachine().getStdout();

            short commandAddr;
            short commandData;
            short value = 0;

            for (int ctr = 0; ctr < ADDR_CARDINALITY; ctr++)
            {
                if (bitString.charAt(8 + ctr) == '1')
                    machineAddrLight[ctr].setImageIndex(
                        ToyImageManager.FRAME_SIM_LIGHT_ON
                        );
                else
                    machineAddrLight[ctr].setImageIndex(
                        ToyImageManager.FRAME_SIM_LIGHT_OFF
                        );
            }


            if (stdout.length > 0)
                value = stdout[stdout.length - 1].getValue();

            for (int ctr = 0; ctr < OUTPUT_CARDINALITY; ctr++)
            {
                if (stdout.length > 0)
                    machineOutputLCD[ctr].setImageIndex(
                        ToyImageManager.FRAME_SIM_DIGIT +
                            ((value >> (12 - ctr * 4)) & 0x000F)
                        );
                else
                    machineOutputLCD[ctr].setImageIndex(
                        ToyImageManager.FRAME_SIM_DIGIT_BLANK
                        );
            }

            bitString = workspace.getVirtualMachine().getCurrentMem().toBinaryString();

            for (int ctr = 0; ctr < DATA_CARDINALITY; ctr++)
            {
                if (bitString.charAt(ctr) == '1')
                    machineDataLight[ctr].setImageIndex(
                        ToyImageManager.FRAME_SIM_LIGHT_ON
                        );
                else
                    machineDataLight[ctr].setImageIndex(
                        ToyImageManager.FRAME_SIM_LIGHT_OFF
                        );
            }


            if (viewChoice.getSelectedIndex() == 0)
            {
                viewScrollbar.setMaximum(
                    (int)Math.ceil(1.0 * ToyVirtualMachine.MEM_CARDINALITY / VIEW_ROWS)
                    );

                viewScrollbar.setVisibleAmount(VIEW_COLUMNS);

                if (!supressScrollFix)
                    viewScrollbar.setValue(
                        workspace.getVirtualMachine().getProgramCtr() / VIEW_ROWS
                        );

                for (int ctr = 0; ctr < VIEW_COLUMNS * VIEW_ROWS; ctr++)
                {
                    viewLabel[ctr].setText(
                        ToyWord.toHexString((short)(viewScrollbar.getValue() * VIEW_ROWS + ctr)).substring(2) +
                        ":" +
                        workspace.getVirtualMachine().getMem(viewScrollbar.getValue() * VIEW_ROWS + ctr).toHexString()
                        );

                    viewLabel[ctr].setForeground(new Color(0, 0, 0));

                    if (viewScrollbar.getValue() * VIEW_ROWS + ctr == workspace.getVirtualMachine().getProgramCtr())
                        viewLabel[ctr].setBackground(new Color(153, 171, 187));
                    else
                        viewLabel[ctr].setBackground(new Color(193, 193, 193));
                }
            }

            if (viewChoice.getSelectedIndex() == 1)
            {
                viewScrollbar.setMaximum(
                    (int)Math.ceil((workspace.getVirtualMachine().getStdin().length + 1.0) / VIEW_ROWS)
                    );

                viewScrollbar.setVisibleAmount(VIEW_COLUMNS);

                if (!supressScrollFix)
                    viewScrollbar.setValue(
                        workspace.getVirtualMachine().getStdinCtr() / VIEW_ROWS
                        );

                for (int ctr = 0; ctr < VIEW_COLUMNS * VIEW_ROWS; ctr++)
                {
                    if (viewScrollbar.getValue() * VIEW_ROWS + ctr < stdin.length)
                        viewLabel[ctr].setText(
                            stdin[viewScrollbar.getValue() * VIEW_ROWS + ctr].toHexString()
                            );
                    else {
                        if (viewScrollbar.getValue() * VIEW_ROWS + ctr == stdin.length)
                            viewLabel[ctr].setText("(Prompt)");
                        else
                            viewLabel[ctr].setText("-");

                    }

                    if (viewScrollbar.getValue() * VIEW_ROWS + ctr < workspace.getVirtualMachine().getStdinCtr())
                        viewLabel[ctr].setForeground(new Color(255, 0, 0));
                    else
                        viewLabel[ctr].setForeground(new Color(0, 0, 0));

                    if (viewScrollbar.getValue() * VIEW_ROWS + ctr == workspace.getVirtualMachine().getStdinCtr())
                        viewLabel[ctr].setBackground(new Color(153, 171, 187));
                    else
                        viewLabel[ctr].setBackground(new Color(193, 193, 193));
                }
            }

            if (viewChoice.getSelectedIndex() == 2)
            {
                viewScrollbar.setMaximum(
                    (int)Math.ceil(1.0 * workspace.getVirtualMachine().getStdout().length / VIEW_ROWS)
                    );

                viewScrollbar.setVisibleAmount(VIEW_COLUMNS);

                if (!supressScrollFix)
                    viewScrollbar.setValue(
                        (workspace.getVirtualMachine().getStdout().length - 1) / VIEW_ROWS
                        );

                for (int ctr = 0; ctr < VIEW_COLUMNS * VIEW_ROWS; ctr++)
                {
                    if (viewScrollbar.getValue() * VIEW_ROWS + ctr < stdout.length)
                        viewLabel[ctr].setText(
                            stdout[viewScrollbar.getValue() * VIEW_ROWS + ctr].toHexString()
                            );
                    else
                        viewLabel[ctr].setText("-");

                    viewLabel[ctr].setForeground(new Color(0, 0, 0));

                    if (viewScrollbar.getValue() * VIEW_ROWS + ctr == stdout.length - 1)
                        viewLabel[ctr].setBackground(new Color(153, 171, 187));
                    else
                        viewLabel[ctr].setBackground(new Color(193, 193, 193));
                }
            }

            for (short ctr = 0; ctr < ToyVirtualMachine.REGISTER_CARDINALITY; ctr++)
                registerContent[ctr].setText(
                    workspace.getVirtualMachine().getRegister(ctr).toHexString()
                    );

            registerCurrentHex.setText(
                ToyWord.toHexString(
                    workspace.getVirtualMachine().getProgramCtr()
                    ).substring(2) + ":" +
                workspace.getVirtualMachine().getCurrentMem().toHexString()
                );

            registerCurrentPseudocode.setText(
                workspace.getVirtualMachine().getCurrentMem().toPseudoCodeString()
                );

            commandAddr = 0;
            for (int ctr = 0; ctr < 8; ctr++)
            {
                commandAddr <<= 1;
                if (machineAddrSwitch[ctr].getState())
                    commandAddr |= 1;
            }

            commandData = 0;
            for (int ctr = 0; ctr < 16; ctr++)
            {
                commandData <<= 1;
                if (machineDataSwitch[ctr].getState())
                    commandData |= 1;
            }

            registerCommandHex.setText(
                ToyWord.toHexString(commandAddr).substring(2) +
                ":" +
                ToyWord.toHexString(commandData)
                );

            registerCommandPseudocode.setText(
                ToyWord.toPseudoCodeString(commandData)
                );

        }

        repaint();
        Graphics g = getGraphics();
        g.translate(viewScrollbar.getBounds().x, viewScrollbar.getBounds().y);
        viewScrollbar.paint(g);
    }

    //--------------------------------------------------------------------------
    //Implement ActionListener
    //--------------------------------------------------------------------------
    public void actionPerformed(ActionEvent e)
    {
        if (e.getSource() == machineLookButton)
        {
            short commandAddr = 0;
            for (int ctr = 0; ctr < 8; ctr++)
            {
                commandAddr <<= 1;
                if (machineAddrSwitch[ctr].getState())
                    commandAddr |= 1;
            }

            workspace.getVirtualMachine().setProgramCtr(commandAddr);
        }

        if (e.getSource() == machineLoadButton)
        {
            short commandAddr = 0, commandData = 0;
            for (int ctr = 0; ctr < 8; ctr++)
            {
                commandAddr <<= 1;
                if (machineAddrSwitch[ctr].getState())
                    commandAddr |= 1;
            }

            for (int ctr = 0; ctr < 16; ctr++)
            {
                commandData <<= 1;
                if (machineDataSwitch[ctr].getState())
                    commandData |= 1;
            }

            workspace.getVirtualMachine().setProgramCtr(commandAddr);

            workspace.getInitialVirtualMachine().setMem(
                commandAddr,
                new ToyWord(commandData)
                );

            workspace.getVirtualMachine().setMem(
                commandAddr,
                new ToyWord(commandData)
                );
        }

        if (e.getSource() == machineStepButton ||
            e.getSource() == stepMenuItem && isVisible())
        {
            workspace.getVirtualMachine().step(false, false, false);
            refresh();
            if (workspace.getVirtualMachine().requiresInput())
                ToyPopup.askForInput(this, workspace.getVirtualMachine());
        }

        if (e.getSource() == machineRunButton ||
            e.getSource() == runMenuItem && isVisible())
        {
            ToyPopup.execute(
                this,
                this,
                imageManager,
                workspace.getVirtualMachine(),
                closeExecutionCheckbox
                );
        }

        if (e.getSource() == machineResetButton ||
            e.getSource() == resetMenuItem && isVisible())
            workspace.reset();

        refresh();
    }

    //--------------------------------------------------------------------------
    //Implement AdjustmentListener
    //--------------------------------------------------------------------------
    public void adjustmentValueChanged(AdjustmentEvent e)
    {
        refresh(true);
    }

    //--------------------------------------------------------------------------
    //Implement ItemListener
    //--------------------------------------------------------------------------
    public void itemStateChanged(ItemEvent e)
    {
        if (e.getSource() == viewChoice)
        {
            if (viewChoice.getSelectedIndex() == 0)
            {
                viewScrollbar.setMaximum(
                    (int)Math.ceil(1.0 * ToyVirtualMachine.MEM_CARDINALITY / VIEW_ROWS)
                    );

                viewScrollbar.setVisibleAmount(VIEW_COLUMNS);

                viewScrollbar.setValue(
                    workspace.getVirtualMachine().getProgramCtr() / VIEW_ROWS
                    );
            }
            if (viewChoice.getSelectedIndex() == 1)
            {
                viewScrollbar.setMaximum(
                    (int)Math.ceil((workspace.getVirtualMachine().getStdin().length + 1.0) / VIEW_ROWS)
                    );

                viewScrollbar.setVisibleAmount(VIEW_COLUMNS);

                viewScrollbar.setValue(
                    workspace.getVirtualMachine().getStdinCtr() / VIEW_ROWS
                    );
            }

            if (viewChoice.getSelectedIndex() == 2)
            {
                viewScrollbar.setMaximum(
                    (int)Math.ceil(1.0 * workspace.getVirtualMachine().getStdout().length / VIEW_ROWS)
                    );

                viewScrollbar.setVisibleAmount(VIEW_COLUMNS);

                viewScrollbar.setValue(
                    (workspace.getVirtualMachine().getStdout().length - 1) / VIEW_ROWS
                    );
            }
        }

        refresh();
    }
}