package toy.dialog;

import java.awt.*;
import java.awt.event.*;

import toy.*;
import toy.image.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyTitlePanel extends Panel
                           implements ItemListener, ItemSelectable
{
    private ItemListener itemListener;

    private ToyFocusManager focusManager;
    private ToyImageManager imageManager;

    private ToyImageToggleButton editButton;
    private ToyImageToggleButton debugButton;
    private ToyImageToggleButton simButton;

    public ToyTitlePanel(ToyFocusManager newFocusManager,
                       ToyImageManager newImageManager)
    {
        super();

        setLayout(new GridBagLayout());

        focusManager = newFocusManager;
        imageManager = newImageManager;

        setForeground(new Color(29, 48, 57));
        setBackground(new Color(52, 64, 76));

        add(
            new ToySpacer(0, 25),
            ToyToolkit.createGridBagConstraints(
                0, 0,
                1, 1,
                1000.0, 1.0,
                GridBagConstraints.NORTHWEST,
                GridBagConstraints.BOTH,
                new Insets(0, 0, 0, 0),
                0, 0
                )
            );


        editButton = new ToyImageToggleButton(
            imageManager,
            ToyImageManager.FRAME_TITLE_EDIT_OFF,
            ToyImageManager.FRAME_TITLE_EDIT_OFF,
            ToyImageManager.FRAME_TITLE_EDIT_HOVER,
            ToyImageManager.FRAME_TITLE_EDIT_ON,
            ToyImageManager.FRAME_TITLE_EDIT_ON,
            ToyImageManager.FRAME_TITLE_EDIT_ON,
            ToyImageManager.FRAME_TITLE_EDIT_ON
            );
        editButton.setState(true);
        editButton.addItemListener(this);
        editButton.addFocusListener(focusManager);
        add(
            editButton,
            ToyToolkit.createGridBagConstraints(
                1, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.SOUTHEAST,
                GridBagConstraints.NONE,
                new Insets(0, 0, 0, 5),
                0, 0
                )
            );

        debugButton = new ToyImageToggleButton(
            imageManager,
            ToyImageManager.FRAME_TITLE_DEBUG_OFF,
            ToyImageManager.FRAME_TITLE_DEBUG_OFF,
            ToyImageManager.FRAME_TITLE_DEBUG_HOVER,
            ToyImageManager.FRAME_TITLE_DEBUG_ON,
            ToyImageManager.FRAME_TITLE_DEBUG_ON,
            ToyImageManager.FRAME_TITLE_DEBUG_ON,
            ToyImageManager.FRAME_TITLE_DEBUG_ON
            );
        debugButton.addItemListener(this);
        debugButton.addFocusListener(focusManager);
        add(
            debugButton,
            ToyToolkit.createGridBagConstraints(
                2, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.SOUTHEAST,
                GridBagConstraints.NONE,
                new Insets(0, 0, 0, 5),
                0, 0
                )
            );

        simButton = new ToyImageToggleButton(
            imageManager,
            ToyImageManager.FRAME_TITLE_SIM_OFF,
            ToyImageManager.FRAME_TITLE_SIM_OFF,
            ToyImageManager.FRAME_TITLE_SIM_HOVER,
            ToyImageManager.FRAME_TITLE_SIM_ON,
            ToyImageManager.FRAME_TITLE_SIM_ON,
            ToyImageManager.FRAME_TITLE_SIM_ON,
            ToyImageManager.FRAME_TITLE_SIM_ON
            );
        simButton.addItemListener(this);
        simButton.addFocusListener(focusManager);
        add(
            simButton,
            ToyToolkit.createGridBagConstraints(
                3, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.SOUTHEAST,
                GridBagConstraints.NONE,
                new Insets(0, 0, 0, 5),
                0, 0
                )
            );
    }

    ////////////////////////////////////////////////////////////////////////////
    // The get and set functions
    public int getSelectedIndex()
    {
        if (editButton.getState())
            return ToyFrame.EDIT;
        if (debugButton.getState())
            return ToyFrame.DEBUG;
        if (simButton.getState())
            return ToyFrame.SIM;

        //ok something's wrong, pretend we're in edit mode
        return ToyFrame.NULL;
    }

    public void setSelectedIndex(int newState)
    {
        if (newState == ToyFrame.EDIT ||
            newState == ToyFrame.DEBUG ||
            newState == ToyFrame.SIM ||
            newState == ToyFrame.NULL)
        {
            //update the labels and buttons based on the mode (at the top)
            if (newState == ToyFrame.EDIT)
                editButton.requestFocus();

            if (newState == ToyFrame.DEBUG)
                debugButton.requestFocus();

            if (newState == ToyFrame.SIM)
                simButton.requestFocus();

            editButton.setEnabled(newState != ToyFrame.NULL);
            debugButton.setEnabled(newState != ToyFrame.NULL);
            simButton.setEnabled(newState != ToyFrame.NULL);

            editButton.setState(newState == ToyFrame.EDIT);
            debugButton.setState(newState == ToyFrame.DEBUG);
            simButton.setState(newState == ToyFrame.SIM);
        }
    }

    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        super.setBounds(x, y, width, height);

        doLayout();
        repaint();
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ItemSelectable

    public void addItemListener(ItemListener newItemListener)
    {
        itemListener = AWTEventMulticaster.add(itemListener, newItemListener);
    }

    public void removeItemListener(ItemListener newItemListener)
    {
       itemListener = AWTEventMulticaster.remove(itemListener, newItemListener);
    }

    public Object[] getSelectedObjects()
    {
        Object answer[] = new Object[1];

        if (editButton.getState())
            answer[0] = editButton;
        if (debugButton.getState())
            answer[0] = debugButton;
        if (simButton.getState())
            answer[0] = simButton;

        return answer;
    }

    ////////////////////////////////////////////////////////////////////////////
    // Override the paint functions
    public void update(Graphics g)
    {
        paint(g);
    }

    public void paint(Graphics g)
    {
        if (g == null || !isVisible())
            return;

        g.setColor(getBackground());
        g.fillRect(0, 0, getBounds().width, getBounds().height);

        g.setColor(getForeground());
        g.drawLine(0, getBounds().height - 1, getBounds().width, getBounds().height - 1);

        //paint children
        super.paint(g);
    }


    ////////////////////////////////////////////////////////////////////////////
    // Implement ItemListener

    public void itemStateChanged(ItemEvent e)
    {
        if (e.getSource() == editButton && itemListener != null)
            itemListener.itemStateChanged(
                new ItemEvent(
                    this,
                    ItemEvent.ITEM_STATE_CHANGED,
                    new Integer(ToyFrame.EDIT),
                    ItemEvent.SELECTED
                    )
                );

        if (e.getSource() == debugButton && itemListener != null)
            itemListener.itemStateChanged(
                new ItemEvent(
                    this,
                    ItemEvent.ITEM_STATE_CHANGED,
                    new Integer(ToyFrame.DEBUG),
                    ItemEvent.SELECTED
                    )
                );

        if (e.getSource() == simButton && itemListener != null)
            itemListener.itemStateChanged(
                new ItemEvent(
                    this,
                    ItemEvent.ITEM_STATE_CHANGED,
                    new Integer(ToyFrame.SIM),
                    ItemEvent.SELECTED
                    )
                );
    }
}