package edu.princeton.swing;

import java.awt.*;
import javax.swing.*;

/**
 * PSwingUtilities contains some standard swing library functions for Princeton's swing extensions.
 *
 * @author btsang
 * @version 7.1
 */
public abstract class PSwingUtilities
{
    /**
     * An instance of the default toolkit.
     */
    public static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    /**
     * Causes the nearest enclosing validate-root to be validated immediately.
     */
    public static void revalidateImmediately(Component component)
    {
        while (component != null &&
               !(component instanceof JComponent && ((JComponent)component).isValidateRoot()))
            component = component.getParent();

        if (component != null)
            component.validate();
    }

    /**
     * Paints the window containing this component immediately.
     */
    public static void paintWindowImmediately(Component component)
    {
        while (component != null && !(component instanceof Window))
            component = component.getParent();

        if (component != null)
        {
            if (component instanceof JComponent)
            {
                ((JComponent)component).paintImmediately(
                    0,
                    0,
                    component.getWidth(),
                    component.getHeight()
                    );
            }
            else {
                component.paint(component.getGraphics());
            }
        }
    }

}
